# TODO: maybe some OpenRAVE links are disabled
# http://openrave.org/docs/0.8.2/collada_robot_extensions/

# < extra
# type = "collision" >
# < technique
# profile = "OpenRAVE"
# ignore_link_pair

NEVER_COLLISIONS = [('base_bellow_link', 'base_footprint'), ('base_bellow_link', 'bl_caster_l_wheel_link'),
                    ('base_bellow_link', 'bl_caster_r_wheel_link'), ('base_bellow_link', 'bl_caster_rotation_link'),
                    ('base_bellow_link', 'br_caster_l_wheel_link'), ('base_bellow_link', 'br_caster_r_wheel_link'),
                    ('base_bellow_link', 'br_caster_rotation_link'), ('base_bellow_link', 'double_stereo_link'),
                    ('base_bellow_link', 'fl_caster_l_wheel_link'), ('base_bellow_link', 'fl_caster_r_wheel_link'),
                    ('base_bellow_link', 'fl_caster_rotation_link'), ('base_bellow_link', 'fr_caster_l_wheel_link'),
                    ('base_bellow_link', 'fr_caster_r_wheel_link'), ('base_bellow_link', 'fr_caster_rotation_link'),
                    ('base_bellow_link', 'head_pan_link'), ('base_bellow_link', 'head_plate_frame'),
                    ('base_bellow_link', 'head_tilt_link'), ('base_bellow_link', 'l_elbow_flex_link'),
                    ('base_bellow_link', 'l_forearm_link'), ('base_bellow_link', 'l_forearm_roll_link'),
                    ('base_bellow_link', 'l_gripper_motor_accelerometer_link'),
                    ('base_bellow_link', 'l_shoulder_lift_link'), ('base_bellow_link', 'l_shoulder_pan_link'),
                    ('base_bellow_link', 'l_upper_arm_link'), ('base_bellow_link', 'l_upper_arm_roll_link'),
                    ('base_bellow_link', 'l_wrist_roll_link'), ('base_bellow_link', 'laser_tilt_mount_link'),
                    ('base_bellow_link', 'r_elbow_flex_link'), ('base_bellow_link', 'r_forearm_link'),
                    ('base_bellow_link', 'r_forearm_roll_link'),
                    ('base_bellow_link', 'r_gripper_motor_accelerometer_link'),
                    ('base_bellow_link', 'r_shoulder_lift_link'), ('base_bellow_link', 'r_shoulder_pan_link'),
                    ('base_bellow_link', 'r_upper_arm_link'), ('base_bellow_link', 'r_upper_arm_roll_link'),
                    ('base_bellow_link', 'r_wrist_flex_link'), ('base_bellow_link', 'r_wrist_roll_link'),
                    ('base_bellow_link', 'sensor_mount_link'), ('base_footprint', 'bl_caster_l_wheel_link'),
                    ('base_footprint', 'bl_caster_r_wheel_link'), ('base_footprint', 'bl_caster_rotation_link'),
                    ('base_footprint', 'br_caster_l_wheel_link'), ('base_footprint', 'br_caster_r_wheel_link'),
                    ('base_footprint', 'br_caster_rotation_link'), ('base_footprint', 'double_stereo_link'),
                    ('base_footprint', 'fl_caster_l_wheel_link'), ('base_footprint', 'fl_caster_r_wheel_link'),
                    ('base_footprint', 'fl_caster_rotation_link'), ('base_footprint', 'fr_caster_l_wheel_link'),
                    ('base_footprint', 'fr_caster_r_wheel_link'), ('base_footprint', 'fr_caster_rotation_link'),
                    ('base_footprint', 'head_pan_link'), ('base_footprint', 'head_plate_frame'),
                    ('base_footprint', 'head_tilt_link'), ('base_footprint', 'l_elbow_flex_link'),
                    ('base_footprint', 'l_forearm_link'), ('base_footprint', 'l_forearm_roll_link'),
                    ('base_footprint', 'l_gripper_l_finger_link'), ('base_footprint', 'l_gripper_l_finger_tip_link'),
                    ('base_footprint', 'l_gripper_motor_accelerometer_link'), ('base_footprint', 'l_gripper_palm_link'),
                    ('base_footprint', 'l_gripper_r_finger_link'), ('base_footprint', 'l_gripper_r_finger_tip_link'),
                    ('base_footprint', 'l_shoulder_lift_link'), ('base_footprint', 'l_shoulder_pan_link'),
                    ('base_footprint', 'l_upper_arm_link'), ('base_footprint', 'l_upper_arm_roll_link'),
                    ('base_footprint', 'l_wrist_flex_link'), ('base_footprint', 'l_wrist_roll_link'),
                    ('base_footprint', 'laser_tilt_mount_link'), ('base_footprint', 'r_elbow_flex_link'),
                    ('base_footprint', 'r_forearm_link'), ('base_footprint', 'r_forearm_roll_link'),
                    ('base_footprint', 'r_gripper_l_finger_link'), ('base_footprint', 'r_gripper_l_finger_tip_link'),
                    ('base_footprint', 'r_gripper_motor_accelerometer_link'), ('base_footprint', 'r_gripper_palm_link'),
                    ('base_footprint', 'r_gripper_r_finger_link'), ('base_footprint', 'r_gripper_r_finger_tip_link'),
                    ('base_footprint', 'r_shoulder_lift_link'), ('base_footprint', 'r_shoulder_pan_link'),
                    ('base_footprint', 'r_upper_arm_link'), ('base_footprint', 'r_upper_arm_roll_link'),
                    ('base_footprint', 'r_wrist_flex_link'), ('base_footprint', 'r_wrist_roll_link'),
                    ('base_footprint', 'sensor_mount_link'), ('base_footprint', 'torso_lift_link'),
                    ('base_link', 'double_stereo_link'), ('base_link', 'head_pan_link'),
                    ('base_link', 'head_plate_frame'), ('base_link', 'head_tilt_link'),
                    ('base_link', 'l_shoulder_lift_link'), ('base_link', 'l_shoulder_pan_link'),
                    ('base_link', 'l_upper_arm_link'), ('base_link', 'l_upper_arm_roll_link'),
                    ('base_link', 'laser_tilt_mount_link'), ('base_link', 'r_shoulder_lift_link'),
                    ('base_link', 'r_shoulder_pan_link'), ('base_link', 'r_upper_arm_link'),
                    ('base_link', 'r_upper_arm_roll_link'), ('base_link', 'sensor_mount_link'),
                    ('bl_caster_l_wheel_link', 'bl_caster_r_wheel_link'),
                    ('bl_caster_l_wheel_link', 'br_caster_l_wheel_link'),
                    ('bl_caster_l_wheel_link', 'br_caster_r_wheel_link'),
                    ('bl_caster_l_wheel_link', 'br_caster_rotation_link'),
                    ('bl_caster_l_wheel_link', 'double_stereo_link'),
                    ('bl_caster_l_wheel_link', 'fl_caster_l_wheel_link'),
                    ('bl_caster_l_wheel_link', 'fl_caster_r_wheel_link'),
                    ('bl_caster_l_wheel_link', 'fl_caster_rotation_link'),
                    ('bl_caster_l_wheel_link', 'fr_caster_l_wheel_link'),
                    ('bl_caster_l_wheel_link', 'fr_caster_r_wheel_link'),
                    ('bl_caster_l_wheel_link', 'fr_caster_rotation_link'), ('bl_caster_l_wheel_link', 'head_pan_link'),
                    ('bl_caster_l_wheel_link', 'head_plate_frame'), ('bl_caster_l_wheel_link', 'head_tilt_link'),
                    ('bl_caster_l_wheel_link', 'l_elbow_flex_link'), ('bl_caster_l_wheel_link', 'l_forearm_link'),
                    ('bl_caster_l_wheel_link', 'l_forearm_roll_link'),
                    ('bl_caster_l_wheel_link', 'l_gripper_motor_accelerometer_link'),
                    ('bl_caster_l_wheel_link', 'l_shoulder_lift_link'),
                    ('bl_caster_l_wheel_link', 'l_shoulder_pan_link'), ('bl_caster_l_wheel_link', 'l_upper_arm_link'),
                    ('bl_caster_l_wheel_link', 'l_upper_arm_roll_link'),
                    ('bl_caster_l_wheel_link', 'l_wrist_roll_link'),
                    ('bl_caster_l_wheel_link', 'laser_tilt_mount_link'),
                    ('bl_caster_l_wheel_link', 'r_elbow_flex_link'), ('bl_caster_l_wheel_link', 'r_forearm_link'),
                    ('bl_caster_l_wheel_link', 'r_forearm_roll_link'),
                    ('bl_caster_l_wheel_link', 'r_gripper_l_finger_link'),
                    ('bl_caster_l_wheel_link', 'r_gripper_l_finger_tip_link'),
                    ('bl_caster_l_wheel_link', 'r_gripper_motor_accelerometer_link'),
                    ('bl_caster_l_wheel_link', 'r_gripper_palm_link'),
                    ('bl_caster_l_wheel_link', 'r_gripper_r_finger_link'),
                    ('bl_caster_l_wheel_link', 'r_gripper_r_finger_tip_link'),
                    ('bl_caster_l_wheel_link', 'r_shoulder_lift_link'),
                    ('bl_caster_l_wheel_link', 'r_shoulder_pan_link'), ('bl_caster_l_wheel_link', 'r_upper_arm_link'),
                    ('bl_caster_l_wheel_link', 'r_upper_arm_roll_link'),
                    ('bl_caster_l_wheel_link', 'r_wrist_flex_link'), ('bl_caster_l_wheel_link', 'r_wrist_roll_link'),
                    ('bl_caster_l_wheel_link', 'sensor_mount_link'), ('bl_caster_l_wheel_link', 'torso_lift_link'),
                    ('bl_caster_r_wheel_link', 'br_caster_l_wheel_link'),
                    ('bl_caster_r_wheel_link', 'br_caster_r_wheel_link'),
                    ('bl_caster_r_wheel_link', 'br_caster_rotation_link'),
                    ('bl_caster_r_wheel_link', 'double_stereo_link'),
                    ('bl_caster_r_wheel_link', 'fl_caster_l_wheel_link'),
                    ('bl_caster_r_wheel_link', 'fl_caster_r_wheel_link'),
                    ('bl_caster_r_wheel_link', 'fl_caster_rotation_link'),
                    ('bl_caster_r_wheel_link', 'fr_caster_l_wheel_link'),
                    ('bl_caster_r_wheel_link', 'fr_caster_r_wheel_link'),
                    ('bl_caster_r_wheel_link', 'fr_caster_rotation_link'), ('bl_caster_r_wheel_link', 'head_pan_link'),
                    ('bl_caster_r_wheel_link', 'head_plate_frame'), ('bl_caster_r_wheel_link', 'head_tilt_link'),
                    ('bl_caster_r_wheel_link', 'l_elbow_flex_link'), ('bl_caster_r_wheel_link', 'l_forearm_roll_link'),
                    ('bl_caster_r_wheel_link', 'l_gripper_motor_accelerometer_link'),
                    ('bl_caster_r_wheel_link', 'l_shoulder_lift_link'),
                    ('bl_caster_r_wheel_link', 'l_shoulder_pan_link'), ('bl_caster_r_wheel_link', 'l_upper_arm_link'),
                    ('bl_caster_r_wheel_link', 'l_upper_arm_roll_link'),
                    ('bl_caster_r_wheel_link', 'laser_tilt_mount_link'),
                    ('bl_caster_r_wheel_link', 'r_elbow_flex_link'), ('bl_caster_r_wheel_link', 'r_forearm_link'),
                    ('bl_caster_r_wheel_link', 'r_forearm_roll_link'),
                    ('bl_caster_r_wheel_link', 'r_gripper_l_finger_link'),
                    ('bl_caster_r_wheel_link', 'r_gripper_l_finger_tip_link'),
                    ('bl_caster_r_wheel_link', 'r_gripper_motor_accelerometer_link'),
                    ('bl_caster_r_wheel_link', 'r_gripper_palm_link'),
                    ('bl_caster_r_wheel_link', 'r_gripper_r_finger_link'),
                    ('bl_caster_r_wheel_link', 'r_gripper_r_finger_tip_link'),
                    ('bl_caster_r_wheel_link', 'r_shoulder_lift_link'),
                    ('bl_caster_r_wheel_link', 'r_shoulder_pan_link'), ('bl_caster_r_wheel_link', 'r_upper_arm_link'),
                    ('bl_caster_r_wheel_link', 'r_upper_arm_roll_link'),
                    ('bl_caster_r_wheel_link', 'r_wrist_flex_link'), ('bl_caster_r_wheel_link', 'r_wrist_roll_link'),
                    ('bl_caster_r_wheel_link', 'sensor_mount_link'), ('bl_caster_r_wheel_link', 'torso_lift_link'),
                    ('bl_caster_rotation_link', 'br_caster_l_wheel_link'),
                    ('bl_caster_rotation_link', 'br_caster_r_wheel_link'),
                    ('bl_caster_rotation_link', 'br_caster_rotation_link'),
                    ('bl_caster_rotation_link', 'double_stereo_link'),
                    ('bl_caster_rotation_link', 'fl_caster_l_wheel_link'),
                    ('bl_caster_rotation_link', 'fl_caster_r_wheel_link'),
                    ('bl_caster_rotation_link', 'fl_caster_rotation_link'),
                    ('bl_caster_rotation_link', 'fr_caster_l_wheel_link'),
                    ('bl_caster_rotation_link', 'fr_caster_r_wheel_link'),
                    ('bl_caster_rotation_link', 'fr_caster_rotation_link'),
                    ('bl_caster_rotation_link', 'head_pan_link'), ('bl_caster_rotation_link', 'head_plate_frame'),
                    ('bl_caster_rotation_link', 'head_tilt_link'), ('bl_caster_rotation_link', 'l_elbow_flex_link'),
                    ('bl_caster_rotation_link', 'l_forearm_roll_link'),
                    ('bl_caster_rotation_link', 'l_shoulder_lift_link'),
                    ('bl_caster_rotation_link', 'l_shoulder_pan_link'), ('bl_caster_rotation_link', 'l_upper_arm_link'),
                    ('bl_caster_rotation_link', 'l_upper_arm_roll_link'),
                    ('bl_caster_rotation_link', 'laser_tilt_mount_link'),
                    ('bl_caster_rotation_link', 'r_elbow_flex_link'), ('bl_caster_rotation_link', 'r_forearm_link'),
                    ('bl_caster_rotation_link', 'r_forearm_roll_link'),
                    ('bl_caster_rotation_link', 'r_gripper_l_finger_link'),
                    ('bl_caster_rotation_link', 'r_gripper_l_finger_tip_link'),
                    ('bl_caster_rotation_link', 'r_gripper_motor_accelerometer_link'),
                    ('bl_caster_rotation_link', 'r_gripper_palm_link'),
                    ('bl_caster_rotation_link', 'r_gripper_r_finger_link'),
                    ('bl_caster_rotation_link', 'r_shoulder_lift_link'),
                    ('bl_caster_rotation_link', 'r_shoulder_pan_link'), ('bl_caster_rotation_link', 'r_upper_arm_link'),
                    ('bl_caster_rotation_link', 'r_upper_arm_roll_link'),
                    ('bl_caster_rotation_link', 'r_wrist_flex_link'), ('bl_caster_rotation_link', 'r_wrist_roll_link'),
                    ('bl_caster_rotation_link', 'sensor_mount_link'), ('bl_caster_rotation_link', 'torso_lift_link'),
                    ('br_caster_l_wheel_link', 'br_caster_r_wheel_link'),
                    ('br_caster_l_wheel_link', 'double_stereo_link'),
                    ('br_caster_l_wheel_link', 'fl_caster_l_wheel_link'),
                    ('br_caster_l_wheel_link', 'fl_caster_r_wheel_link'),
                    ('br_caster_l_wheel_link', 'fl_caster_rotation_link'),
                    ('br_caster_l_wheel_link', 'fr_caster_l_wheel_link'),
                    ('br_caster_l_wheel_link', 'fr_caster_r_wheel_link'),
                    ('br_caster_l_wheel_link', 'fr_caster_rotation_link'), ('br_caster_l_wheel_link', 'head_pan_link'),
                    ('br_caster_l_wheel_link', 'head_plate_frame'), ('br_caster_l_wheel_link', 'head_tilt_link'),
                    ('br_caster_l_wheel_link', 'l_elbow_flex_link'), ('br_caster_l_wheel_link', 'l_forearm_link'),
                    ('br_caster_l_wheel_link', 'l_forearm_roll_link'),
                    ('br_caster_l_wheel_link', 'l_gripper_l_finger_link'),
                    ('br_caster_l_wheel_link', 'l_gripper_l_finger_tip_link'),
                    ('br_caster_l_wheel_link', 'l_gripper_motor_accelerometer_link'),
                    ('br_caster_l_wheel_link', 'l_gripper_palm_link'),
                    ('br_caster_l_wheel_link', 'l_gripper_r_finger_link'),
                    ('br_caster_l_wheel_link', 'l_gripper_r_finger_tip_link'),
                    ('br_caster_l_wheel_link', 'l_shoulder_lift_link'),
                    ('br_caster_l_wheel_link', 'l_shoulder_pan_link'), ('br_caster_l_wheel_link', 'l_upper_arm_link'),
                    ('br_caster_l_wheel_link', 'l_upper_arm_roll_link'),
                    ('br_caster_l_wheel_link', 'l_wrist_flex_link'), ('br_caster_l_wheel_link', 'l_wrist_roll_link'),
                    ('br_caster_l_wheel_link', 'laser_tilt_mount_link'),
                    ('br_caster_l_wheel_link', 'r_elbow_flex_link'), ('br_caster_l_wheel_link', 'r_forearm_roll_link'),
                    ('br_caster_l_wheel_link', 'r_gripper_motor_accelerometer_link'),
                    ('br_caster_l_wheel_link', 'r_shoulder_lift_link'),
                    ('br_caster_l_wheel_link', 'r_shoulder_pan_link'), ('br_caster_l_wheel_link', 'r_upper_arm_link'),
                    ('br_caster_l_wheel_link', 'r_upper_arm_roll_link'),
                    ('br_caster_l_wheel_link', 'sensor_mount_link'), ('br_caster_l_wheel_link', 'torso_lift_link'),
                    ('br_caster_r_wheel_link', 'double_stereo_link'),
                    ('br_caster_r_wheel_link', 'fl_caster_l_wheel_link'),
                    ('br_caster_r_wheel_link', 'fl_caster_r_wheel_link'),
                    ('br_caster_r_wheel_link', 'fl_caster_rotation_link'),
                    ('br_caster_r_wheel_link', 'fr_caster_l_wheel_link'),
                    ('br_caster_r_wheel_link', 'fr_caster_r_wheel_link'),
                    ('br_caster_r_wheel_link', 'fr_caster_rotation_link'), ('br_caster_r_wheel_link', 'head_pan_link'),
                    ('br_caster_r_wheel_link', 'head_plate_frame'), ('br_caster_r_wheel_link', 'head_tilt_link'),
                    ('br_caster_r_wheel_link', 'l_elbow_flex_link'), ('br_caster_r_wheel_link', 'l_forearm_link'),
                    ('br_caster_r_wheel_link', 'l_forearm_roll_link'),
                    ('br_caster_r_wheel_link', 'l_gripper_l_finger_link'),
                    ('br_caster_r_wheel_link', 'l_gripper_l_finger_tip_link'),
                    ('br_caster_r_wheel_link', 'l_gripper_motor_accelerometer_link'),
                    ('br_caster_r_wheel_link', 'l_gripper_palm_link'),
                    ('br_caster_r_wheel_link', 'l_gripper_r_finger_link'),
                    ('br_caster_r_wheel_link', 'l_gripper_r_finger_tip_link'),
                    ('br_caster_r_wheel_link', 'l_shoulder_lift_link'),
                    ('br_caster_r_wheel_link', 'l_shoulder_pan_link'), ('br_caster_r_wheel_link', 'l_upper_arm_link'),
                    ('br_caster_r_wheel_link', 'l_upper_arm_roll_link'),
                    ('br_caster_r_wheel_link', 'l_wrist_flex_link'), ('br_caster_r_wheel_link', 'l_wrist_roll_link'),
                    ('br_caster_r_wheel_link', 'laser_tilt_mount_link'),
                    ('br_caster_r_wheel_link', 'r_elbow_flex_link'), ('br_caster_r_wheel_link', 'r_forearm_roll_link'),
                    ('br_caster_r_wheel_link', 'r_gripper_motor_accelerometer_link'),
                    ('br_caster_r_wheel_link', 'r_shoulder_lift_link'),
                    ('br_caster_r_wheel_link', 'r_shoulder_pan_link'), ('br_caster_r_wheel_link', 'r_upper_arm_link'),
                    ('br_caster_r_wheel_link', 'r_upper_arm_roll_link'),
                    ('br_caster_r_wheel_link', 'sensor_mount_link'), ('br_caster_r_wheel_link', 'torso_lift_link'),
                    ('br_caster_rotation_link', 'double_stereo_link'),
                    ('br_caster_rotation_link', 'fl_caster_l_wheel_link'),
                    ('br_caster_rotation_link', 'fl_caster_r_wheel_link'),
                    ('br_caster_rotation_link', 'fl_caster_rotation_link'),
                    ('br_caster_rotation_link', 'fr_caster_l_wheel_link'),
                    ('br_caster_rotation_link', 'fr_caster_r_wheel_link'),
                    ('br_caster_rotation_link', 'fr_caster_rotation_link'),
                    ('br_caster_rotation_link', 'head_pan_link'), ('br_caster_rotation_link', 'head_plate_frame'),
                    ('br_caster_rotation_link', 'head_tilt_link'), ('br_caster_rotation_link', 'l_elbow_flex_link'),
                    ('br_caster_rotation_link', 'l_forearm_link'), ('br_caster_rotation_link', 'l_forearm_roll_link'),
                    ('br_caster_rotation_link', 'l_gripper_motor_accelerometer_link'),
                    ('br_caster_rotation_link', 'l_gripper_palm_link'),
                    ('br_caster_rotation_link', 'l_gripper_r_finger_link'),
                    ('br_caster_rotation_link', 'l_gripper_r_finger_tip_link'),
                    ('br_caster_rotation_link', 'l_shoulder_lift_link'),
                    ('br_caster_rotation_link', 'l_shoulder_pan_link'), ('br_caster_rotation_link', 'l_upper_arm_link'),
                    ('br_caster_rotation_link', 'l_upper_arm_roll_link'),
                    ('br_caster_rotation_link', 'l_wrist_flex_link'), ('br_caster_rotation_link', 'l_wrist_roll_link'),
                    ('br_caster_rotation_link', 'laser_tilt_mount_link'),
                    ('br_caster_rotation_link', 'r_elbow_flex_link'),
                    ('br_caster_rotation_link', 'r_forearm_roll_link'),
                    ('br_caster_rotation_link', 'r_gripper_motor_accelerometer_link'),
                    ('br_caster_rotation_link', 'r_shoulder_lift_link'),
                    ('br_caster_rotation_link', 'r_shoulder_pan_link'), ('br_caster_rotation_link', 'r_upper_arm_link'),
                    ('br_caster_rotation_link', 'r_upper_arm_roll_link'),
                    ('br_caster_rotation_link', 'sensor_mount_link'), ('br_caster_rotation_link', 'torso_lift_link'),
                    ('double_stereo_link', 'fl_caster_l_wheel_link'), ('double_stereo_link', 'fl_caster_r_wheel_link'),
                    ('double_stereo_link', 'fl_caster_rotation_link'), ('double_stereo_link', 'fr_caster_l_wheel_link'),
                    ('double_stereo_link', 'fr_caster_r_wheel_link'), ('double_stereo_link', 'fr_caster_rotation_link'),
                    ('double_stereo_link', 'l_elbow_flex_link'), ('double_stereo_link', 'l_forearm_link'),
                    ('double_stereo_link', 'l_forearm_roll_link'), ('double_stereo_link', 'l_gripper_l_finger_link'),
                    ('double_stereo_link', 'l_gripper_l_finger_tip_link'),
                    ('double_stereo_link', 'l_gripper_motor_accelerometer_link'),
                    ('double_stereo_link', 'l_shoulder_lift_link'), ('double_stereo_link', 'l_shoulder_pan_link'),
                    ('double_stereo_link', 'l_upper_arm_link'), ('double_stereo_link', 'l_upper_arm_roll_link'),
                    ('double_stereo_link', 'l_wrist_flex_link'), ('double_stereo_link', 'l_wrist_roll_link'),
                    ('double_stereo_link', 'laser_tilt_mount_link'), ('double_stereo_link', 'r_elbow_flex_link'),
                    ('double_stereo_link', 'r_forearm_link'), ('double_stereo_link', 'r_forearm_roll_link'),
                    ('double_stereo_link', 'r_gripper_l_finger_link'),
                    ('double_stereo_link', 'r_gripper_l_finger_tip_link'),
                    ('double_stereo_link', 'r_gripper_motor_accelerometer_link'),
                    ('double_stereo_link', 'r_gripper_palm_link'), ('double_stereo_link', 'r_gripper_r_finger_link'),
                    ('double_stereo_link', 'r_gripper_r_finger_tip_link'),
                    ('double_stereo_link', 'r_shoulder_lift_link'), ('double_stereo_link', 'r_shoulder_pan_link'),
                    ('double_stereo_link', 'r_upper_arm_link'), ('double_stereo_link', 'r_upper_arm_roll_link'),
                    ('double_stereo_link', 'r_wrist_flex_link'), ('double_stereo_link', 'r_wrist_roll_link'),
                    ('double_stereo_link', 'torso_lift_link'), ('fl_caster_l_wheel_link', 'fl_caster_r_wheel_link'),
                    ('fl_caster_l_wheel_link', 'fr_caster_l_wheel_link'),
                    ('fl_caster_l_wheel_link', 'fr_caster_r_wheel_link'),
                    ('fl_caster_l_wheel_link', 'fr_caster_rotation_link'), ('fl_caster_l_wheel_link', 'head_pan_link'),
                    ('fl_caster_l_wheel_link', 'head_plate_frame'), ('fl_caster_l_wheel_link', 'head_tilt_link'),
                    ('fl_caster_l_wheel_link', 'l_elbow_flex_link'), ('fl_caster_l_wheel_link', 'l_forearm_roll_link'),
                    ('fl_caster_l_wheel_link', 'l_shoulder_lift_link'),
                    ('fl_caster_l_wheel_link', 'l_shoulder_pan_link'), ('fl_caster_l_wheel_link', 'l_upper_arm_link'),
                    ('fl_caster_l_wheel_link', 'l_upper_arm_roll_link'),
                    ('fl_caster_l_wheel_link', 'laser_tilt_mount_link'),
                    ('fl_caster_l_wheel_link', 'r_elbow_flex_link'), ('fl_caster_l_wheel_link', 'r_forearm_link'),
                    ('fl_caster_l_wheel_link', 'r_forearm_roll_link'),
                    ('fl_caster_l_wheel_link', 'r_gripper_motor_accelerometer_link'),
                    ('fl_caster_l_wheel_link', 'r_shoulder_lift_link'),
                    ('fl_caster_l_wheel_link', 'r_shoulder_pan_link'), ('fl_caster_l_wheel_link', 'r_upper_arm_link'),
                    ('fl_caster_l_wheel_link', 'r_upper_arm_roll_link'),
                    ('fl_caster_l_wheel_link', 'r_wrist_flex_link'), ('fl_caster_l_wheel_link', 'r_wrist_roll_link'),
                    ('fl_caster_l_wheel_link', 'sensor_mount_link'), ('fl_caster_l_wheel_link', 'torso_lift_link'),
                    ('fl_caster_r_wheel_link', 'fr_caster_l_wheel_link'),
                    ('fl_caster_r_wheel_link', 'fr_caster_r_wheel_link'),
                    ('fl_caster_r_wheel_link', 'fr_caster_rotation_link'), ('fl_caster_r_wheel_link', 'head_pan_link'),
                    ('fl_caster_r_wheel_link', 'head_plate_frame'), ('fl_caster_r_wheel_link', 'head_tilt_link'),
                    ('fl_caster_r_wheel_link', 'l_elbow_flex_link'), ('fl_caster_r_wheel_link', 'l_forearm_roll_link'),
                    ('fl_caster_r_wheel_link', 'l_shoulder_lift_link'),
                    ('fl_caster_r_wheel_link', 'l_shoulder_pan_link'), ('fl_caster_r_wheel_link', 'l_upper_arm_link'),
                    ('fl_caster_r_wheel_link', 'l_upper_arm_roll_link'),
                    ('fl_caster_r_wheel_link', 'laser_tilt_mount_link'),
                    ('fl_caster_r_wheel_link', 'r_elbow_flex_link'), ('fl_caster_r_wheel_link', 'r_forearm_link'),
                    ('fl_caster_r_wheel_link', 'r_forearm_roll_link'),
                    ('fl_caster_r_wheel_link', 'r_gripper_motor_accelerometer_link'),
                    ('fl_caster_r_wheel_link', 'r_shoulder_lift_link'),
                    ('fl_caster_r_wheel_link', 'r_shoulder_pan_link'), ('fl_caster_r_wheel_link', 'r_upper_arm_link'),
                    ('fl_caster_r_wheel_link', 'r_upper_arm_roll_link'),
                    ('fl_caster_r_wheel_link', 'r_wrist_flex_link'), ('fl_caster_r_wheel_link', 'r_wrist_roll_link'),
                    ('fl_caster_r_wheel_link', 'sensor_mount_link'), ('fl_caster_r_wheel_link', 'torso_lift_link'),
                    ('fl_caster_rotation_link', 'fr_caster_l_wheel_link'),
                    ('fl_caster_rotation_link', 'fr_caster_r_wheel_link'),
                    ('fl_caster_rotation_link', 'fr_caster_rotation_link'),
                    ('fl_caster_rotation_link', 'head_pan_link'), ('fl_caster_rotation_link', 'head_plate_frame'),
                    ('fl_caster_rotation_link', 'head_tilt_link'), ('fl_caster_rotation_link', 'l_elbow_flex_link'),
                    ('fl_caster_rotation_link', 'l_forearm_roll_link'),
                    ('fl_caster_rotation_link', 'l_shoulder_lift_link'),
                    ('fl_caster_rotation_link', 'l_shoulder_pan_link'), ('fl_caster_rotation_link', 'l_upper_arm_link'),
                    ('fl_caster_rotation_link', 'l_upper_arm_roll_link'),
                    ('fl_caster_rotation_link', 'laser_tilt_mount_link'),
                    ('fl_caster_rotation_link', 'r_elbow_flex_link'),
                    ('fl_caster_rotation_link', 'r_forearm_roll_link'),
                    ('fl_caster_rotation_link', 'r_gripper_motor_accelerometer_link'),
                    ('fl_caster_rotation_link', 'r_shoulder_lift_link'),
                    ('fl_caster_rotation_link', 'r_shoulder_pan_link'), ('fl_caster_rotation_link', 'r_upper_arm_link'),
                    ('fl_caster_rotation_link', 'r_upper_arm_roll_link'),
                    ('fl_caster_rotation_link', 'sensor_mount_link'), ('fl_caster_rotation_link', 'torso_lift_link'),
                    ('fr_caster_l_wheel_link', 'fr_caster_r_wheel_link'), ('fr_caster_l_wheel_link', 'head_pan_link'),
                    ('fr_caster_l_wheel_link', 'head_plate_frame'), ('fr_caster_l_wheel_link', 'head_tilt_link'),
                    ('fr_caster_l_wheel_link', 'l_elbow_flex_link'), ('fr_caster_l_wheel_link', 'l_forearm_link'),
                    ('fr_caster_l_wheel_link', 'l_forearm_roll_link'),
                    ('fr_caster_l_wheel_link', 'l_gripper_motor_accelerometer_link'),
                    ('fr_caster_l_wheel_link', 'l_shoulder_lift_link'),
                    ('fr_caster_l_wheel_link', 'l_shoulder_pan_link'), ('fr_caster_l_wheel_link', 'l_upper_arm_link'),
                    ('fr_caster_l_wheel_link', 'l_upper_arm_roll_link'),
                    ('fr_caster_l_wheel_link', 'l_wrist_flex_link'), ('fr_caster_l_wheel_link', 'l_wrist_roll_link'),
                    ('fr_caster_l_wheel_link', 'laser_tilt_mount_link'),
                    ('fr_caster_l_wheel_link', 'r_elbow_flex_link'), ('fr_caster_l_wheel_link', 'r_forearm_roll_link'),
                    ('fr_caster_l_wheel_link', 'r_shoulder_lift_link'),
                    ('fr_caster_l_wheel_link', 'r_shoulder_pan_link'), ('fr_caster_l_wheel_link', 'r_upper_arm_link'),
                    ('fr_caster_l_wheel_link', 'r_upper_arm_roll_link'),
                    ('fr_caster_l_wheel_link', 'sensor_mount_link'), ('fr_caster_l_wheel_link', 'torso_lift_link'),
                    ('fr_caster_r_wheel_link', 'head_pan_link'), ('fr_caster_r_wheel_link', 'head_plate_frame'),
                    ('fr_caster_r_wheel_link', 'head_tilt_link'), ('fr_caster_r_wheel_link', 'l_elbow_flex_link'),
                    ('fr_caster_r_wheel_link', 'l_forearm_link'), ('fr_caster_r_wheel_link', 'l_forearm_roll_link'),
                    ('fr_caster_r_wheel_link', 'l_gripper_motor_accelerometer_link'),
                    ('fr_caster_r_wheel_link', 'l_shoulder_lift_link'),
                    ('fr_caster_r_wheel_link', 'l_shoulder_pan_link'), ('fr_caster_r_wheel_link', 'l_upper_arm_link'),
                    ('fr_caster_r_wheel_link', 'l_upper_arm_roll_link'),
                    ('fr_caster_r_wheel_link', 'l_wrist_flex_link'), ('fr_caster_r_wheel_link', 'l_wrist_roll_link'),
                    ('fr_caster_r_wheel_link', 'laser_tilt_mount_link'),
                    ('fr_caster_r_wheel_link', 'r_elbow_flex_link'), ('fr_caster_r_wheel_link', 'r_forearm_roll_link'),
                    ('fr_caster_r_wheel_link', 'r_shoulder_lift_link'),
                    ('fr_caster_r_wheel_link', 'r_shoulder_pan_link'), ('fr_caster_r_wheel_link', 'r_upper_arm_link'),
                    ('fr_caster_r_wheel_link', 'r_upper_arm_roll_link'),
                    ('fr_caster_r_wheel_link', 'sensor_mount_link'), ('fr_caster_r_wheel_link', 'torso_lift_link'),
                    ('fr_caster_rotation_link', 'head_pan_link'), ('fr_caster_rotation_link', 'head_plate_frame'),
                    ('fr_caster_rotation_link', 'head_tilt_link'), ('fr_caster_rotation_link', 'l_elbow_flex_link'),
                    ('fr_caster_rotation_link', 'l_forearm_roll_link'),
                    ('fr_caster_rotation_link', 'l_gripper_motor_accelerometer_link'),
                    ('fr_caster_rotation_link', 'l_shoulder_lift_link'),
                    ('fr_caster_rotation_link', 'l_shoulder_pan_link'), ('fr_caster_rotation_link', 'l_upper_arm_link'),
                    ('fr_caster_rotation_link', 'l_upper_arm_roll_link'),
                    ('fr_caster_rotation_link', 'laser_tilt_mount_link'),
                    ('fr_caster_rotation_link', 'r_elbow_flex_link'),
                    ('fr_caster_rotation_link', 'r_forearm_roll_link'),
                    ('fr_caster_rotation_link', 'r_shoulder_lift_link'),
                    ('fr_caster_rotation_link', 'r_shoulder_pan_link'), ('fr_caster_rotation_link', 'r_upper_arm_link'),
                    ('fr_caster_rotation_link', 'r_upper_arm_roll_link'),
                    ('fr_caster_rotation_link', 'sensor_mount_link'), ('fr_caster_rotation_link', 'torso_lift_link'),
                    ('head_pan_link', 'l_elbow_flex_link'), ('head_pan_link', 'l_forearm_roll_link'),
                    ('head_pan_link', 'l_gripper_motor_accelerometer_link'), ('head_pan_link', 'l_shoulder_lift_link'),
                    ('head_pan_link', 'l_shoulder_pan_link'), ('head_pan_link', 'l_upper_arm_link'),
                    ('head_pan_link', 'l_upper_arm_roll_link'), ('head_pan_link', 'laser_tilt_mount_link'),
                    ('head_pan_link', 'r_elbow_flex_link'), ('head_pan_link', 'r_forearm_roll_link'),
                    ('head_pan_link', 'r_gripper_motor_accelerometer_link'), ('head_pan_link', 'r_shoulder_lift_link'),
                    ('head_pan_link', 'r_shoulder_pan_link'), ('head_pan_link', 'r_upper_arm_link'),
                    ('head_pan_link', 'r_upper_arm_roll_link'), ('head_pan_link', 'r_wrist_roll_link'),
                    ('head_plate_frame', 'l_elbow_flex_link'), ('head_plate_frame', 'l_forearm_link'),
                    ('head_plate_frame', 'l_forearm_roll_link'), ('head_plate_frame', 'l_gripper_l_finger_link'),
                    ('head_plate_frame', 'l_gripper_l_finger_tip_link'),
                    ('head_plate_frame', 'l_gripper_motor_accelerometer_link'),
                    ('head_plate_frame', 'l_gripper_palm_link'), ('head_plate_frame', 'l_gripper_r_finger_link'),
                    ('head_plate_frame', 'l_gripper_r_finger_tip_link'), ('head_plate_frame', 'l_shoulder_lift_link'),
                    ('head_plate_frame', 'l_shoulder_pan_link'), ('head_plate_frame', 'l_upper_arm_link'),
                    ('head_plate_frame', 'l_upper_arm_roll_link'), ('head_plate_frame', 'l_wrist_flex_link'),
                    ('head_plate_frame', 'l_wrist_roll_link'), ('head_plate_frame', 'laser_tilt_mount_link'),
                    ('head_plate_frame', 'r_elbow_flex_link'), ('head_plate_frame', 'r_forearm_link'),
                    ('head_plate_frame', 'r_forearm_roll_link'), ('head_plate_frame', 'r_gripper_l_finger_link'),
                    ('head_plate_frame', 'r_gripper_l_finger_tip_link'),
                    ('head_plate_frame', 'r_gripper_motor_accelerometer_link'),
                    ('head_plate_frame', 'r_gripper_palm_link'), ('head_plate_frame', 'r_gripper_r_finger_link'),
                    ('head_plate_frame', 'r_gripper_r_finger_tip_link'), ('head_plate_frame', 'r_shoulder_lift_link'),
                    ('head_plate_frame', 'r_shoulder_pan_link'), ('head_plate_frame', 'r_upper_arm_link'),
                    ('head_plate_frame', 'r_upper_arm_roll_link'), ('head_plate_frame', 'r_wrist_flex_link'),
                    ('head_plate_frame', 'r_wrist_roll_link'), ('head_plate_frame', 'torso_lift_link'),
                    ('head_tilt_link', 'l_elbow_flex_link'), ('head_tilt_link', 'l_forearm_roll_link'),
                    ('head_tilt_link', 'l_gripper_motor_accelerometer_link'),
                    ('head_tilt_link', 'l_shoulder_lift_link'), ('head_tilt_link', 'l_shoulder_pan_link'),
                    ('head_tilt_link', 'l_upper_arm_link'), ('head_tilt_link', 'l_upper_arm_roll_link'),
                    ('head_tilt_link', 'laser_tilt_mount_link'), ('head_tilt_link', 'r_elbow_flex_link'),
                    ('head_tilt_link', 'r_forearm_roll_link'), ('head_tilt_link', 'r_gripper_motor_accelerometer_link'),
                    ('head_tilt_link', 'r_shoulder_lift_link'), ('head_tilt_link', 'r_shoulder_pan_link'),
                    ('head_tilt_link', 'r_upper_arm_link'), ('head_tilt_link', 'r_upper_arm_roll_link'),
                    ('head_tilt_link', 'r_wrist_roll_link'), ('head_tilt_link', 'torso_lift_link'),
                    ('l_elbow_flex_link', 'l_forearm_link'), ('l_elbow_flex_link', 'l_gripper_l_finger_link'),
                    ('l_elbow_flex_link', 'l_gripper_l_finger_tip_link'),
                    ('l_elbow_flex_link', 'l_gripper_motor_accelerometer_link'),
                    ('l_elbow_flex_link', 'l_gripper_palm_link'), ('l_elbow_flex_link', 'l_gripper_r_finger_link'),
                    ('l_elbow_flex_link', 'l_gripper_r_finger_tip_link'), ('l_elbow_flex_link', 'l_shoulder_lift_link'),
                    ('l_elbow_flex_link', 'l_shoulder_pan_link'), ('l_elbow_flex_link', 'l_upper_arm_roll_link'),
                    ('l_elbow_flex_link', 'l_wrist_flex_link'), ('l_elbow_flex_link', 'l_wrist_roll_link'),
                    ('l_elbow_flex_link', 'laser_tilt_mount_link'), ('l_elbow_flex_link', 'r_shoulder_lift_link'),
                    ('l_elbow_flex_link', 'r_shoulder_pan_link'), ('l_elbow_flex_link', 'r_upper_arm_roll_link'),
                    ('l_elbow_flex_link', 'sensor_mount_link'), ('l_elbow_flex_link', 'torso_lift_link'),
                    ('l_forearm_link', 'l_gripper_l_finger_link'), ('l_forearm_link', 'l_gripper_l_finger_tip_link'),
                    ('l_forearm_link', 'l_gripper_motor_accelerometer_link'), ('l_forearm_link', 'l_gripper_palm_link'),
                    ('l_forearm_link', 'l_gripper_r_finger_link'), ('l_forearm_link', 'l_gripper_r_finger_tip_link'),
                    ('l_forearm_link', 'l_shoulder_lift_link'), ('l_forearm_link', 'l_upper_arm_link'),
                    ('l_forearm_link', 'l_upper_arm_roll_link'), ('l_forearm_link', 'l_wrist_roll_link'),
                    ('l_forearm_link', 'sensor_mount_link'), ('l_forearm_roll_link', 'l_gripper_l_finger_link'),
                    ('l_forearm_roll_link', 'l_gripper_l_finger_tip_link'),
                    ('l_forearm_roll_link', 'l_gripper_motor_accelerometer_link'),
                    ('l_forearm_roll_link', 'l_gripper_palm_link'), ('l_forearm_roll_link', 'l_gripper_r_finger_link'),
                    ('l_forearm_roll_link', 'l_gripper_r_finger_tip_link'),
                    ('l_forearm_roll_link', 'l_shoulder_lift_link'), ('l_forearm_roll_link', 'l_shoulder_pan_link'),
                    ('l_forearm_roll_link', 'l_upper_arm_link'), ('l_forearm_roll_link', 'l_upper_arm_roll_link'),
                    ('l_forearm_roll_link', 'l_wrist_flex_link'), ('l_forearm_roll_link', 'l_wrist_roll_link'),
                    ('l_forearm_roll_link', 'laser_tilt_mount_link'), ('l_forearm_roll_link', 'r_shoulder_lift_link'),
                    ('l_forearm_roll_link', 'r_shoulder_pan_link'), ('l_forearm_roll_link', 'r_upper_arm_roll_link'),
                    ('l_forearm_roll_link', 'sensor_mount_link'), ('l_forearm_roll_link', 'torso_lift_link'),
                    ('l_gripper_l_finger_link', 'l_gripper_motor_accelerometer_link'),
                    ('l_gripper_l_finger_link', 'l_gripper_r_finger_link'),
                    ('l_gripper_l_finger_link', 'l_gripper_r_finger_tip_link'),
                    ('l_gripper_l_finger_link', 'l_shoulder_lift_link'),
                    ('l_gripper_l_finger_link', 'l_upper_arm_link'),
                    ('l_gripper_l_finger_link', 'l_upper_arm_roll_link'),
                    ('l_gripper_l_finger_link', 'l_wrist_flex_link'), ('l_gripper_l_finger_link', 'l_wrist_roll_link'),
                    ('l_gripper_l_finger_link', 'sensor_mount_link'),
                    ('l_gripper_l_finger_tip_link', 'l_gripper_motor_accelerometer_link'),
                    ('l_gripper_l_finger_tip_link', 'l_gripper_palm_link'),
                    ('l_gripper_l_finger_tip_link', 'l_gripper_r_finger_link'),
                    ('l_gripper_l_finger_tip_link', 'l_shoulder_lift_link'),
                    ('l_gripper_l_finger_tip_link', 'l_upper_arm_link'),
                    ('l_gripper_l_finger_tip_link', 'l_upper_arm_roll_link'),
                    ('l_gripper_l_finger_tip_link', 'l_wrist_flex_link'),
                    ('l_gripper_l_finger_tip_link', 'l_wrist_roll_link'),
                    ('l_gripper_l_finger_tip_link', 'sensor_mount_link'),
                    ('l_gripper_motor_accelerometer_link', 'l_gripper_r_finger_link'),
                    ('l_gripper_motor_accelerometer_link', 'l_gripper_r_finger_tip_link'),
                    ('l_gripper_motor_accelerometer_link', 'l_shoulder_lift_link'),
                    ('l_gripper_motor_accelerometer_link', 'l_upper_arm_link'),
                    ('l_gripper_motor_accelerometer_link', 'l_upper_arm_roll_link'),
                    ('l_gripper_motor_accelerometer_link', 'l_wrist_flex_link'),
                    ('l_gripper_motor_accelerometer_link', 'l_wrist_roll_link'),
                    ('l_gripper_motor_accelerometer_link', 'laser_tilt_mount_link'),
                    ('l_gripper_motor_accelerometer_link', 'r_gripper_l_finger_tip_link'),
                    ('l_gripper_motor_accelerometer_link', 'r_gripper_motor_accelerometer_link'),
                    ('l_gripper_motor_accelerometer_link', 'r_gripper_r_finger_tip_link'),
                    ('l_gripper_motor_accelerometer_link', 'r_wrist_flex_link'),
                    ('l_gripper_motor_accelerometer_link', 'r_wrist_roll_link'),
                    ('l_gripper_motor_accelerometer_link', 'sensor_mount_link'),
                    ('l_gripper_palm_link', 'l_gripper_r_finger_tip_link'),
                    ('l_gripper_palm_link', 'l_shoulder_lift_link'), ('l_gripper_palm_link', 'l_upper_arm_link'),
                    ('l_gripper_palm_link', 'l_upper_arm_roll_link'), ('l_gripper_palm_link', 'l_wrist_flex_link'),
                    ('l_gripper_palm_link', 'sensor_mount_link'), ('l_gripper_r_finger_link', 'l_shoulder_lift_link'),
                    ('l_gripper_r_finger_link', 'l_upper_arm_link'),
                    ('l_gripper_r_finger_link', 'l_upper_arm_roll_link'),
                    ('l_gripper_r_finger_link', 'l_wrist_flex_link'), ('l_gripper_r_finger_link', 'l_wrist_roll_link'),
                    ('l_gripper_r_finger_link', 'sensor_mount_link'),
                    ('l_gripper_r_finger_tip_link', 'l_shoulder_lift_link'),
                    ('l_gripper_r_finger_tip_link', 'l_upper_arm_link'),
                    ('l_gripper_r_finger_tip_link', 'l_upper_arm_roll_link'),
                    ('l_gripper_r_finger_tip_link', 'l_wrist_flex_link'),
                    ('l_gripper_r_finger_tip_link', 'l_wrist_roll_link'),
                    ('l_gripper_r_finger_tip_link', 'r_gripper_motor_accelerometer_link'),
                    ('l_gripper_r_finger_tip_link', 'sensor_mount_link'), ('l_shoulder_lift_link', 'l_upper_arm_link'),
                    ('l_shoulder_lift_link', 'l_wrist_flex_link'), ('l_shoulder_lift_link', 'l_wrist_roll_link'),
                    ('l_shoulder_lift_link', 'laser_tilt_mount_link'), ('l_shoulder_lift_link', 'r_elbow_flex_link'),
                    ('l_shoulder_lift_link', 'r_forearm_roll_link'), ('l_shoulder_lift_link', 'r_shoulder_lift_link'),
                    ('l_shoulder_lift_link', 'r_upper_arm_link'), ('l_shoulder_lift_link', 'r_upper_arm_roll_link'),
                    ('l_shoulder_lift_link', 'sensor_mount_link'), ('l_shoulder_lift_link', 'torso_lift_link'),
                    ('l_shoulder_pan_link', 'l_upper_arm_roll_link'), ('l_shoulder_pan_link', 'laser_tilt_mount_link'),
                    ('l_shoulder_pan_link', 'r_elbow_flex_link'), ('l_shoulder_pan_link', 'r_forearm_roll_link'),
                    ('l_shoulder_pan_link', 'sensor_mount_link'), ('l_upper_arm_link', 'l_wrist_flex_link'),
                    ('l_upper_arm_link', 'l_wrist_roll_link'), ('l_upper_arm_link', 'laser_tilt_mount_link'),
                    ('l_upper_arm_link', 'r_shoulder_lift_link'), ('l_upper_arm_link', 'sensor_mount_link'),
                    ('l_upper_arm_link', 'torso_lift_link'), ('l_upper_arm_roll_link', 'l_wrist_flex_link'),
                    ('l_upper_arm_roll_link', 'l_wrist_roll_link'), ('l_upper_arm_roll_link', 'laser_tilt_mount_link'),
                    ('l_upper_arm_roll_link', 'r_elbow_flex_link'), ('l_upper_arm_roll_link', 'r_forearm_roll_link'),
                    ('l_upper_arm_roll_link', 'r_gripper_motor_accelerometer_link'),
                    ('l_upper_arm_roll_link', 'r_shoulder_lift_link'), ('l_upper_arm_roll_link', 'r_upper_arm_link'),
                    ('l_upper_arm_roll_link', 'r_upper_arm_roll_link'), ('l_upper_arm_roll_link', 'sensor_mount_link'),
                    ('l_upper_arm_roll_link', 'torso_lift_link'), ('l_wrist_flex_link', 'sensor_mount_link'),
                    ('l_wrist_roll_link', 'laser_tilt_mount_link'), ('l_wrist_roll_link', 'sensor_mount_link'),
                    ('laser_tilt_mount_link', 'r_elbow_flex_link'), ('laser_tilt_mount_link', 'r_forearm_roll_link'),
                    ('laser_tilt_mount_link', 'r_gripper_motor_accelerometer_link'),
                    ('laser_tilt_mount_link', 'r_shoulder_lift_link'), ('laser_tilt_mount_link', 'r_shoulder_pan_link'),
                    ('laser_tilt_mount_link', 'r_upper_arm_link'), ('laser_tilt_mount_link', 'r_upper_arm_roll_link'),
                    ('laser_tilt_mount_link', 'r_wrist_roll_link'), ('laser_tilt_mount_link', 'sensor_mount_link'),
                    ('r_elbow_flex_link', 'r_forearm_link'), ('r_elbow_flex_link', 'r_gripper_l_finger_link'),
                    ('r_elbow_flex_link', 'r_gripper_l_finger_tip_link'),
                    ('r_elbow_flex_link', 'r_gripper_motor_accelerometer_link'),
                    ('r_elbow_flex_link', 'r_gripper_palm_link'), ('r_elbow_flex_link', 'r_gripper_r_finger_link'),
                    ('r_elbow_flex_link', 'r_gripper_r_finger_tip_link'), ('r_elbow_flex_link', 'r_shoulder_lift_link'),
                    ('r_elbow_flex_link', 'r_shoulder_pan_link'), ('r_elbow_flex_link', 'r_upper_arm_roll_link'),
                    ('r_elbow_flex_link', 'r_wrist_flex_link'), ('r_elbow_flex_link', 'r_wrist_roll_link'),
                    ('r_elbow_flex_link', 'sensor_mount_link'), ('r_elbow_flex_link', 'torso_lift_link'),
                    ('r_forearm_link', 'r_gripper_l_finger_link'), ('r_forearm_link', 'r_gripper_l_finger_tip_link'),
                    ('r_forearm_link', 'r_gripper_motor_accelerometer_link'), ('r_forearm_link', 'r_gripper_palm_link'),
                    ('r_forearm_link', 'r_gripper_r_finger_link'), ('r_forearm_link', 'r_gripper_r_finger_tip_link'),
                    ('r_forearm_link', 'r_shoulder_lift_link'), ('r_forearm_link', 'r_upper_arm_link'),
                    ('r_forearm_link', 'r_upper_arm_roll_link'), ('r_forearm_link', 'r_wrist_roll_link'),
                    ('r_forearm_link', 'sensor_mount_link'), ('r_forearm_roll_link', 'r_gripper_l_finger_link'),
                    ('r_forearm_roll_link', 'r_gripper_l_finger_tip_link'),
                    ('r_forearm_roll_link', 'r_gripper_motor_accelerometer_link'),
                    ('r_forearm_roll_link', 'r_gripper_palm_link'), ('r_forearm_roll_link', 'r_gripper_r_finger_link'),
                    ('r_forearm_roll_link', 'r_gripper_r_finger_tip_link'),
                    ('r_forearm_roll_link', 'r_shoulder_lift_link'), ('r_forearm_roll_link', 'r_shoulder_pan_link'),
                    ('r_forearm_roll_link', 'r_upper_arm_link'), ('r_forearm_roll_link', 'r_upper_arm_roll_link'),
                    ('r_forearm_roll_link', 'r_wrist_flex_link'), ('r_forearm_roll_link', 'r_wrist_roll_link'),
                    ('r_forearm_roll_link', 'sensor_mount_link'), ('r_forearm_roll_link', 'torso_lift_link'),
                    ('r_gripper_l_finger_link', 'r_gripper_motor_accelerometer_link'),
                    ('r_gripper_l_finger_link', 'r_gripper_r_finger_link'),
                    ('r_gripper_l_finger_link', 'r_gripper_r_finger_tip_link'),
                    ('r_gripper_l_finger_link', 'r_shoulder_lift_link'),
                    ('r_gripper_l_finger_link', 'r_upper_arm_link'),
                    ('r_gripper_l_finger_link', 'r_upper_arm_roll_link'),
                    ('r_gripper_l_finger_link', 'r_wrist_flex_link'), ('r_gripper_l_finger_link', 'r_wrist_roll_link'),
                    ('r_gripper_l_finger_link', 'sensor_mount_link'),
                    ('r_gripper_l_finger_tip_link', 'r_gripper_motor_accelerometer_link'),
                    ('r_gripper_l_finger_tip_link', 'r_gripper_palm_link'),
                    ('r_gripper_l_finger_tip_link', 'r_gripper_r_finger_link'),
                    ('r_gripper_l_finger_tip_link', 'r_shoulder_lift_link'),
                    ('r_gripper_l_finger_tip_link', 'r_upper_arm_link'),
                    ('r_gripper_l_finger_tip_link', 'r_upper_arm_roll_link'),
                    ('r_gripper_l_finger_tip_link', 'r_wrist_flex_link'),
                    ('r_gripper_l_finger_tip_link', 'r_wrist_roll_link'),
                    ('r_gripper_l_finger_tip_link', 'sensor_mount_link'),
                    ('r_gripper_motor_accelerometer_link', 'r_gripper_r_finger_link'),
                    ('r_gripper_motor_accelerometer_link', 'r_gripper_r_finger_tip_link'),
                    ('r_gripper_motor_accelerometer_link', 'r_shoulder_lift_link'),
                    ('r_gripper_motor_accelerometer_link', 'r_upper_arm_link'),
                    ('r_gripper_motor_accelerometer_link', 'r_upper_arm_roll_link'),
                    ('r_gripper_motor_accelerometer_link', 'r_wrist_flex_link'),
                    ('r_gripper_motor_accelerometer_link', 'r_wrist_roll_link'),
                    ('r_gripper_motor_accelerometer_link', 'sensor_mount_link'),
                    ('r_gripper_palm_link', 'r_gripper_r_finger_tip_link'),
                    ('r_gripper_palm_link', 'r_shoulder_lift_link'), ('r_gripper_palm_link', 'r_upper_arm_link'),
                    ('r_gripper_palm_link', 'r_upper_arm_roll_link'), ('r_gripper_palm_link', 'r_wrist_flex_link'),
                    ('r_gripper_palm_link', 'sensor_mount_link'), ('r_gripper_r_finger_link', 'r_shoulder_lift_link'),
                    ('r_gripper_r_finger_link', 'r_upper_arm_link'),
                    ('r_gripper_r_finger_link', 'r_upper_arm_roll_link'),
                    ('r_gripper_r_finger_link', 'r_wrist_flex_link'), ('r_gripper_r_finger_link', 'r_wrist_roll_link'),
                    ('r_gripper_r_finger_link', 'sensor_mount_link'),
                    ('r_gripper_r_finger_tip_link', 'r_shoulder_lift_link'),
                    ('r_gripper_r_finger_tip_link', 'r_upper_arm_link'),
                    ('r_gripper_r_finger_tip_link', 'r_upper_arm_roll_link'),
                    ('r_gripper_r_finger_tip_link', 'r_wrist_flex_link'),
                    ('r_gripper_r_finger_tip_link', 'r_wrist_roll_link'),
                    ('r_gripper_r_finger_tip_link', 'sensor_mount_link'), ('r_shoulder_lift_link', 'r_upper_arm_link'),
                    ('r_shoulder_lift_link', 'r_wrist_flex_link'), ('r_shoulder_lift_link', 'r_wrist_roll_link'),
                    ('r_shoulder_lift_link', 'sensor_mount_link'), ('r_shoulder_lift_link', 'torso_lift_link'),
                    ('r_shoulder_pan_link', 'r_upper_arm_roll_link'), ('r_shoulder_pan_link', 'sensor_mount_link'),
                    ('r_upper_arm_link', 'r_wrist_flex_link'), ('r_upper_arm_link', 'r_wrist_roll_link'),
                    ('r_upper_arm_link', 'sensor_mount_link'), ('r_upper_arm_link', 'torso_lift_link'),
                    ('r_upper_arm_roll_link', 'r_wrist_flex_link'), ('r_upper_arm_roll_link', 'r_wrist_roll_link'),
                    ('r_upper_arm_roll_link', 'sensor_mount_link'), ('r_upper_arm_roll_link', 'torso_lift_link'),
                    ('r_wrist_flex_link', 'sensor_mount_link'), ('r_wrist_roll_link', 'sensor_mount_link'),
                    ('sensor_mount_link', 'torso_lift_link')
                    ] + [
    ('l_upper_arm_link', 'l_shoulder_pan_link'),
    ('r_upper_arm_link', 'r_shoulder_pan_link'),
]